# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2012 Phoenix87 <phoenix1987@gmail.com>
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import gettext
from gettext import gettext as _
gettext.textdomain('gtumbler')

import gobject
import gtk
import logging
logger = logging.getLogger('gtumbler')

from gtumbler_lib import Window
from gtumbler_lib import pdf
from gtumbler_lib.command import Command
from gtumbler_lib.preferences import preferences

from gtumbler.AboutGtumblerDialog        import AboutGtumblerDialog
from gtumbler.PreferencesGtumblerDialog  import PreferencesGtumblerDialog
from gtumbler.ProgressDialog			 import ProgressDialog

import gtumbler.message as message
import gtumbler.shell as shell
import gtumbler.mimes as mimes
import os
import poppler
import re
import session
import shutil
import threading
import time
import tempfile

Session = session.Session()

def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)

BOX_NAMES = ['artBox', 'bleedBox', 'cropBox', 'mediaBox', 'trimBox']

BOX_WIDGETS = ['top', 'bottom', 'left', 'right']
    
DOC_COLUMN = enum('UID',
                  'FILENAME',
                  'PATH',
                  'MIMETYPE',
                  'AUTHOR',
                  'CREATOR',
                  'PRODUCER',
                  'SUBJECT',
                  'TITLE',
                 )

DOC_CROP = enum('RECT',
                'SUBSET',
                'PAGES',
               )

PAGE_SUBSET = enum('EVEN',
                   'ODD',
                   'BOTH',
                  )

RECT = enum('TOP',
            'BOTTOM',
            'LEFT',
            'RIGHT',
           )

def produce_ranges(is_include, ranges, n_pages):
	if is_include: # Include
		return set(range(n_pages)) \
			   if not ranges \
			   else unite_ranges(ranges)
	else: # Exclude
		return set() \
			   if not ranges \
			   else set(range(n_pages)) - unite_ranges(ranges)

def unite_ranges(text):
	ranges = text.split(',')
	
	result = set()
	
	for interval in ranges:
		ends = interval.split('-')
		result |= set([int(ends[0].strip()) - 1] if len(ends) == 1 \
				  else range(int(ends[0].strip()) - 1, int(ends[1].strip()))) 
				  
	return result


### BEGIN THREAD CLASSES

class ProgressThread(threading.Thread):
	def __init__(self, parent):
		self.parent = parent # Parent progress dialog
		self._stop = threading.Event()
		super(ProgressThread, self).__init__()

	def stop(self):
		self._stop.set()
		
	def has_stopped(self):
		return self._stop.isSet()

class ConcatenateThread(ProgressThread):
	def __init__(self, parent, data):
		self.main, self.dest_file = data
		super(ConcatenateThread, self).__init__(parent)

	def cat2(self, a, b, c):
		pdf_out = pdf.PdfFileWriter()
		with file(a[0], 'rb') as fin1, file(b[0], 'rb') as fin2, file(c, 'wb') as fout:
			pdf_in1 = pdf.PdfFileReader(fin1)
			pdf_in2 = pdf.PdfFileReader(fin2)
			
			self.parent.set_steps(pdf_in1.getNumPages() + pdf_in2.getNumPages())

			gtk.gdk.threads_enter()
			self.parent.set_current_task("Concatenating documents...")
			gtk.gdk.threads_leave()
			
			
			ranges = produce_ranges(*a[1], n_pages = pdf_in1.getNumPages())
			i = -1
			for page in pdf_in1.pages:
				i += 1
				gtk.gdk.threads_enter()
				self.parent.stepit()
				gtk.gdk.threads_leave()
				if self.has_stopped(): return
				if i not in ranges: continue
				pdf_out.addPage(page)

			ranges = produce_ranges(*b[1], n_pages = pdf_in2.getNumPages())
			i = -1
			for page in pdf_in2.pages:
				i += 1
				gtk.gdk.threads_enter()
				self.parent.stepit()
				gtk.gdk.threads_leave()
				if self.has_stopped(): return
				if i not in ranges: continue
				pdf_out.addPage(page)
			
			gtk.gdk.threads_enter()
			self.parent.set_current_task("Writing data to disk...")
			gtk.gdk.threads_leave()

			gtk.gdk.threads_enter()
			self.parent.spinner(True)
			gtk.gdk.threads_leave()
			
			pdf_out.write(fout)

			gtk.gdk.threads_enter()
			self.parent.set_current_task("Concatenating documents...")
			gtk.gdk.threads_leave()

			gtk.gdk.threads_enter()
			self.parent.spinner(False)
			gtk.gdk.threads_leave()

	#~ def catn(self, p, tmp):
		#~ if len(p) == 1: return p[0]
#~ 
		#~ if len(p) % 2:
			#~ odd = tmp.new_file('.pdf')
			#~ self.cat2(p[0], p[1], odd)
			#~ p[1] = odd
			#~ p = p[1:]
#~ 
		#~ i = 0
		#~ fr = self.parent.get_fraction()
		#~ n = len(p) / 2
		#~ while i < n:
			#~ new_file = tmp.new_file('.pdf')
			#~ self.cat2(p[2 * i], p[2 * i + 1], new_file)
			#~ p[i] = new_file
			#~ #gtk.gdk.threads_enter()
			#~ #self.parent.set_fraction(fr + (1-fr)/n * (i+1))
			#~ #gtk.gdk.threads_leave()
			#~ i += 1
#~ 
		#~ return self.catn(p[:len(p) / 2] , tmp)
		
	def catn(self, p, tmp):
		n = len(p)
		while n > 1:
			i = r = n % 2
			n = n >> 1
			fr = self.parent.get_fraction()
			while i < n + r:
				new_file = tmp.new_file('.pdf')
				self.cat2(p[2 * i - r], p[2 * i + 1 - r], new_file)
				p[i] = (new_file, (True, ''))
				i += 1
			n += r
		return p[0][0]

	def run(self):
		tmp = Session.temp_new('cat')
		p = []
		gtk.gdk.threads_enter()
		self.parent.set_current_task("Scanning document list...")
		self.parent.spinner(True)
		gtk.gdk.threads_leave()

		for row in self.main.ui.ls_documents:
			uid, fname, path, mime = row[DOC_COLUMN.UID], row[DOC_COLUMN.FILENAME], row[DOC_COLUMN.PATH], row[DOC_COLUMN.MIMETYPE]
			temp_file = Session.doc_temp.get(row[DOC_COLUMN.UID])
			
			old_file = temp_file if temp_file else "%s/%s" % (path, fname)
			if mime in mimes.CONVERT_MIMES and preferences['chk_convert']: # Must be converted
				new_file = tmp.new_file('.pdf')
				gtk.gdk.threads_enter()
				self.parent.set_current_task('Converting "%s" from PostScript to PDF...' % old_file)
				self.parent.spinner(True)
				gtk.gdk.threads_leave()
				
				comp = '1.%s' % (2 + preferences['cbo_compatibility'])
				shell.ps2pdf(old_file, new_file, comp)
				
				gtk.gdk.threads_enter()
				self.parent.set_current_task('Scanning document list...')
				self.parent.spinner(False)
				gtk.gdk.threads_leave()				
			else:
				new_file = old_file
			
			p.append( (new_file, Session.doc_ranges.get(uid)) )
		
		gtk.gdk.threads_enter()
		self.parent.spinner(False)
		gtk.gdk.threads_leave()
		
		self.parent.set_max(1. / (len(p) - 1))
		os.rename(self.catn(p, tmp), self.dest_file)
		
		Session.temp_delete('cat')
		
		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		
		self.parent.response(gtk.RESPONSE_OK)

class CropThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data[0]
		self.filename = data[1]
		self.temp_file = data[2]
		self.box = data[3]
		self.rect = data[4]
		self.subset = data[5]
		self.pages = data[6]
		super(CropThread, self).__init__(parent)
		
	def run(self):
		gtk.gdk.threads_enter()
		self.parent.set_current_task('Preparing to apply new Bounding Boxes...')
		gtk.gdk.threads_leave()
				
		with file(self.filename, 'rb') as fin, file(self.temp_file, 'wb') as fout:
			pdf_in = pdf.PdfFileReader(fin)
			pdf_out = pdf.PdfFileWriter()
			
			tot = float(pdf_in.getNumPages())
			i = 0
			        
			for page in pdf_in.pages:
				i += 1
				if self.main.ui.rdo_crop_all_pages.get_active() or i - 1 in self.pages:
					if self.subset - (1 - i % 2):
						gtk.gdk.threads_enter()
						self.parent.set_current_task('Applying Bounding Box to page %d out of a total of %d pages...' % (i, int(tot)))
						self.parent.set_fraction(i / tot)
						gtk.gdk.threads_leave()

						self.bbox(self.box, page)
				pdf_out.addPage(page)
					
			gtk.gdk.threads_enter()
			self.parent.set_current_task('Applying changes. Please wait...')
			self.parent.spinner(True)
			gtk.gdk.threads_leave()
			
			pdf_out.write(fout)

			gtk.gdk.threads_enter()
			self.parent.set_current_task('Changes successfully applied')
			self.parent.spinner(False)
			gtk.gdk.threads_leave()

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		
		self.parent.response(gtk.RESPONSE_OK)

	def bbox(self, box, page):
		left, bottom = page.mediaBox.lowerLeft
		right, top = page.mediaBox.upperRight
		w, h = float(right - left), float(top - bottom)
		getattr(page, BOX_NAMES[box]).lowerLeft = (float(Session.boxes[box][RECT.LEFT]),
												   float(Session.boxes[box][RECT.BOTTOM]))
		getattr(page, BOX_NAMES[box]).upperRight = (w - float(Session.boxes[box][RECT.RIGHT]),
													h - float(Session.boxes[box][RECT.TOP]))

		
class DirectoryCrawlerThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data[0]
		self.d = data[1]
		self.isrec = data[2]
		super(DirectoryCrawlerThread, self).__init__(parent)
			   
	def run(self): ## TODO: Implement the Cancel button
		i = 1
		if self.isrec: ## Scan folders recursively
			gtk.gdk.threads_enter()
			self.parent.set_current_task("Scanning directory tree... This may take some time...")
			gtk.gdk.threads_leave()
			toadd = []

			gtk.gdk.threads_enter()
			self.parent.spinner(True)
			gtk.gdk.threads_leave()

			for root, sdirs, files in os.walk(self.d, followlinks=True):
				for f in files:
					toadd.append("%s/%s" % (root, f))
				if self.has_stopped(): return

			gtk.gdk.threads_enter()
			self.parent.spinner(False)
			gtk.gdk.threads_leave()
			
			tot = float(len(toadd))
			self.parent.set_current_task("A total of %d files have been found. Looking for supported documents..." % len(toadd))
			for f in toadd:
				gtk.gdk.threads_enter()
				self.parent.set_fraction(i / tot)
				self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
				self.main.add_file(f)
				gtk.gdk.threads_leave()
				i += 1
				if self.has_stopped(): return

		else: ## Just this folder
			dirs = os.listdir(self.d)
			tot = float(len(dirs))
			gtk.gdk.threads_enter()
			self.parent.set_current_task("A total of %d entries in the selected folder have been found. Looking for supported documents..." % len(dirs))
			gtk.gdk.threads_leave()
			for f in dirs:
				if os.path.isfile("%s/%s" % (self.d, f)):
					gtk.gdk.threads_enter()
					self.parent.set_fraction(i / tot)
					self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
					self.main.add_file("%s/%s" % (self.d, f))
					gtk.gdk.threads_leave()
				i += 1
				if self.has_stopped(): return

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)


class FileListThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data[0]
		self.flist = data[1]
		super(FileListThread, self).__init__(parent)
			   
	def run(self):
		i = 1
		gtk.gdk.threads_enter()
		self.parent.set_current_task("A total of %d have been selected. Looking for supported documents..." % len(self.flist))
		gtk.gdk.threads_leave()
		tot = float(len(self.flist))
		for f in self.flist:
			gtk.gdk.threads_enter()
			self.parent.set_fraction(i / tot)
			self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
			gtk.gdk.threads_leave()
			self.main.add_file(f)
			i += 1
			if self.has_stopped(): return

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)

class OverwriteThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data
		super(OverwriteThread, self).__init__(parent)

	def run(self):
		tot = float(len(self.main.ui.ls_documents))
		i = 0
		for row in self.main.ui.ls_documents:
			uid = row[DOC_COLUMN.UID]
			temp_file = Session.doc_temp.get(uid)
			old_name = '/'.join((row[DOC_COLUMN.PATH], row[DOC_COLUMN.FILENAME]))
			gtk.gdk.threads_enter()
			self.parent.set_current_task('Processing "%s"...' % old_name)
			gtk.gdk.threads_leave()
			if temp_file:
				fin = temp_file
				# Adjust file name for PostScript
				fout = filename = old_name if row[DOC_COLUMN.MIMETYPE] not in mimes.CONVERT_MIMES \
				                           else '.'.join( (os.path.splitext(old_name)[0], 'pdf') )
			else:
				fin = filename = old_name
				fout = Session.temp_new('overwrite').new_file('.pdf')

			gtk.gdk.threads_enter()
			self.parent.set_current_task('Overwriting "%s"...' % old_name)
			gtk.gdk.threads_leave()

			ranges = Session.doc_ranges.get(uid)
			if not ranges:
				if temp_file: os.rename(temp_file, filename)
			else:
				with file(fin, 'rb') as src, file(fout, 'wb') as dest:
					pdf_in = pdf.PdfFileReader(src)
					pdf_out = pdf.PdfFileWriter()
					j = -1
					ranges = produce_ranges(*Session.doc_ranges.get(uid, (True, '')), n_pages = pdf_in.getNumPages())
					for page in pdf_in.pages:
						j += 1
						if j not in ranges: continue
						pdf_out.addPage(page)
						
					pdf_out.write(dest)
					
				if not temp_file: os.rename(fout, fin)
					
			Session.doc_temp[uid] = None

			i += 1
			gtk.gdk.threads_enter()
			self.parent.set_fraction(i / tot)
			gtk.gdk.threads_leave()

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)
		
class SaveAsThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data
		super(SaveAsThread, self).__init__(parent)

	def run(self):
		tot = float(len(self.main.ui.ls_documents))
		n = 0
		
		# Determine numbering
		pattern  = Session.main.ui.txt_pattern.get_text()
		match = re.search('#{([0-9]),([0-9]+?),([0-9]+?)}', pattern)
		if match:
			d, s, i = match.groups()
			d, s, i = int(d), int(s), int(i)
			num_pattern = re.search('(#{%d,%d,%d})' % (d,s,i) , pattern).group(0)

		prototype = pattern.replace('*name*', '')
		prototype = prototype.replace('*ext*', '')
		prototype = prototype.replace('*path*', '')
		prototype = prototype.replace('*title*', '')
		prototype = prototype.replace('*author*', '')

		for row in self.main.ui.ls_documents:
			uid = row[DOC_COLUMN.UID]
			filename = Session.doc_temp.get(uid)
			old_name = '/'.join((row[DOC_COLUMN.PATH], row[DOC_COLUMN.FILENAME]))
			if not filename: filename = old_name

			gtk.gdk.threads_enter()
			self.parent.set_current_task('Processing "%s"...' % old_name)
			gtk.gdk.threads_leave()

			new_name = pattern.replace('*name*', os.path.splitext(row[DOC_COLUMN.FILENAME])[0])
			new_name = new_name.replace('*ext*', os.path.splitext(row[DOC_COLUMN.FILENAME])[1])
			new_name = new_name.replace('*path*', row[DOC_COLUMN.PATH])
			new_name = new_name.replace('*title*', row[DOC_COLUMN.TITLE] if row[DOC_COLUMN.TITLE] else '')
			new_name = new_name.replace('*author*', row[DOC_COLUMN.AUTHOR] if row[DOC_COLUMN.AUTHOR] else '')
			if match:
				new_name = new_name.replace(num_pattern, ("%%0%dd" % d) % s)
				s += i
			else:
				if new_name == prototype: continue
			new_name = '.'.join((new_name, 'pdf'))
			path = Session.main.ui.fcb_dest_folder.get_filename()
			print '/'.join((path,new_name))
			
			ranges = Session.doc_ranges.get(uid)

			gtk.gdk.threads_enter()
			self.parent.set_current_task('Saving "%s" as "%s"...' % (filename, new_name))
			gtk.gdk.threads_leave()

			if ranges:
				temp_file = Session.temp_new('saveas').new_file('.pdf')
				with file(filename, 'rb') as src, file(temp_file, 'wb') as dest:
					pdf_in = pdf.PdfFileReader(src)
					pdf_out = pdf.PdfFileWriter()
					j = -1
					ranges = produce_ranges(*Session.doc_ranges.get(uid, (True, '')), n_pages = pdf_in.getNumPages())
					for page in pdf_in.pages:
						j += 1
						if j not in ranges: continue
						pdf_out.addPage(page)
						
					pdf_out.write(dest)
					
				os.rename(temp_file, '/'.join((path, new_name)))
			else:
				shutil.copyfile(filename, '/'.join((path, new_name)))

			n += 1
			gtk.gdk.threads_enter()
			self.parent.set_fraction(i / tot)
			gtk.gdk.threads_leave()

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)
		
### END THREAD CLASSES

### BEGIN COMMAND CLASSES

class ApplyCrop(Command):
	"""This is the dummiest thing in the World"""
	
	def do(self):
		with ProgressDialog() as pd:
			self._uid, = Session.main.get_from_selected_row(DOC_COLUMN.UID)
			filename = Session.main.get_selected_active_filename()
			temp_file = Session.main.get_selected_active_filename(True)
			rename = False
			if filename == temp_file:
				temp_file = Session.temp_new('crop').new_file('.pdf')
				rename = True
			box = Session.main.ui.cbo_box.get_active()
			rect = Session.boxes[box]
			#TODO: Check for valid syntax in the entrybox
			subset = Session.main.ui.cbo_subset.get_active()

			pages = None if Session.main.ui.rdo_crop_all_pages.get_active() \
					else unite_ranges(Session.main.ui.txt_custom_pages.get_text())

			pd.thread = CropThread(pd, [Session.main, filename, temp_file, box, rect, subset, pages])
			resp = pd.start()
			if resp != gtk.RESPONSE_OK:
				pd.stop()
				return resp
			
			if rename:
				self._prev = filename
				Session.doc_temp[self._uid] = temp_file
			else:
				self._prev = Session.temp_new('undo').new_file('.pdf')
				shutil.copyfile(filename, self._prev)

			return resp

	def undo(self):
		if not getattr(self, '_next', None):
			self._next = Session.doc_temp[self._uid]
		Session.doc_temp[self._uid] = self._prev
		self.preview = DocumentPreview(self._next, Session.main.ui.da_preview)
		Session.main.preview.show_page()
		Session.main.update_bbox()
		
	def redo(self):
		Session.doc_temp[self._uid] = self._next
		self.preview = DocumentPreview(self._prev, Session.main.ui.da_preview)
		Session.main.preview.show_page()
		Session.main.update_bbox()

class ApplyRanges(Command):
	def __init__(self, uid, is_include, ranges):
		self._uid = uid
		self._is_include = is_include
		self._ranges = ranges
		super(ApplyRanges, self).__init__()
		
	def do(self):
		self._prev = Session.doc_ranges.get(self._uid, (True, ''))
		Session.doc_ranges[self._uid] = (self._is_include, self._ranges)
		
	def undo(self):
		if not getattr(self, '_next', None):
			self._next = Session.doc_ranges[self._uid][:]
		Session.doc_ranges[self._uid] = self._prev

	def redo(self):
		Session.doc_ranges[self._uid] = self._next

class DocListCommand(Command):
	def __init__(self, receiver = None, previous = None):
		self._receiver = receiver
		self._prev = previous
		super(DocListCommand, self).__init__()
		
	def do(self):
		if self._prev == None: self._prev = Session._dump_doc_list()
		if self._receiver: self._receiver()
		
	def undo(self):
		self._next = Session._dump_doc_list()
		self._execute()
	
	def redo(self):
		self._next, self._prev = self._prev, self._next
		self._execute()
		self.undo = self.redo

	def _execute(self):
		Session._undump_doc_list(self._prev)


class SetDocInfo(Command):
	_widgets = ['author', 'creator', 'producer', 'subject', 'title']
	
	def do(self):
		self.redo = self.undo
		
		self._next = [getattr(Session.main.ui, 'txt_%s' % widget).get_text() for widget in self._widgets]
		self._prev = Session.main.get_from_selected_row(4, 5, 6, 7, 8)
		self._it = Session.main.ui.tv_documents.get_selection().get_selected()[1]
		self._execute()
	
	def undo(self):
		self._prev, self._next = self._next, self._prev
		self._execute()
		
	def _execute(self):
		i = 4
		for field in self._next:
			Session.main.ui.ls_documents.set_value(self._it, i, field)
			i += 1
			

### END COMMAND CLASSES

### BEGIN OTHER CLASSES

class DocumentPreview(object):
    def __init__(self, filename, drawing_area):
        self.filename = filename
        self.drawing_area = drawing_area
        self.page_number = None     
        self.zoom = 1
        self.pad = 3
        
        self.document = poppler.document_new_from_file("file://%s" % self.filename, None)
        self.pdf_rd = pdf.PdfFileReader(file(filename, 'rb'))

    def expose(self):
        s = self.zoom
        cr = self.drawing_area.window.cairo_create()
		
		# Grey page shadow
		cr.set_source_rgb(.3, .3, .3)
        cr.rectangle(self.pad, self.pad, s * self.width, s * self.height)
        cr.fill()
		
		# White default page background
		cr.set_source_rgb(1, 1, 1)
        cr.rectangle(0, 0, s * self.width, s * self.height)
        cr.fill()
        
        # Black page contour
        cr.set_source_rgb(0, 0, 0)
        cr.set_line_width(.8)
        cr.rectangle(0, 0, s * self.width, s * self.height)
        cr.stroke()
        
        # Rendering the current page
        cr = self.drawing_area.window.cairo_create()
        if s != 1:
            cr.scale(self.zoom, self.zoom)

		cr.translate(*self.origin)
        self.current_page.render(cr)

		self.drawing_area.set_size_request(self.pad + int(self.width * self.zoom),
		                                   self.pad + int(self.height * self.zoom))

    def get_current_page(self):
        return self.page_number

    def get_n_pages(self):
        return self.document.get_n_pages()

    def get_page_size(self):
        return self.width, self.height

    def get_zoom(self):
        return self.zoom

    def get_title(self):
        return self.document.get_property("title")

    def set_zoom(self, zoom):
        self.zoom = zoom
        self.show_page()

    def show_page(self, page = None):
		if page != None and page != self.page_number:
			self.current_page = self.document.get_page(page)
			paper = self.pdf_rd.getPage(page)
			left, bottom = paper.mediaBox.lowerLeft
			right, top = paper.mediaBox.upperRight

			self.width, self.height = float(right - left), float(top - bottom)
			for box_name, i in zip(BOX_NAMES, range(5)):
				box = getattr(paper, box_name, None)
				if box:
					Session.boxes[i] = [self.height - float(box.upperRight[1]),
					                    float(box.lowerLeft[1]),
					                    float(box.lowerLeft[0]),
					                    self.width - float(box.upperRight[0])]

			self.origin = (float(paper.cropBox.lowerLeft[0]), self.height - float(paper.cropBox.upperRight[1]))
			self.page_number = page
		
        self.drawing_area.queue_draw()
        try:
			self.expose()
		except AttributeError:
			pass

	def zoom_in(self, delta):
		self.zoom += delta
		self.show_page()
		
	def zoom_out(self, delta):
		if self.zoom - delta > 0:
			self.zoom -= delta
			self.show_page()

### END OTHER CLASSES

# See gtumbler_lib.Window.py for more details about how this class works
class GtumblerWindow(Window):
	__gtype_name__ = "GtumblerWindow"
	
	def finish_initializing(self, builder): # pylint: disable=E1002
		"""Set up the main window"""
		super(GtumblerWindow, self).finish_initializing(builder)

		self.AboutDialog = AboutGtumblerDialog
		self.PreferencesDialog = PreferencesGtumblerDialog

		# Code for other initialization actions should be added here.
		self.ui.fc_add_folder.add_buttons(gtk.STOCK_OK, gtk.RESPONSE_OK,
										  gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

### BEGIN LOCAL HELPERS

	def add_file(self, f):
		# Trust the MIME Type instead of the extension (which might be
		# missing)
		mime = shell.get_mimetype(f)
		
		# Test for a valid MIME Type
		if mime not in mimes.CONVERT_MIMES:
			if mime not in mimes.CAT_MIMES:
				return
			else:
				with file(f, 'rb') as fin:
					pdf_in = pdf.PdfFileReader(fin)
					doc_info = pdf_in.getDocumentInfo()
				author, creator, producer, subject, title = [doc_info.author, doc_info.creator, doc_info.producer, doc_info.subject, doc_info.title]
		else:
			author, creator, producer, subject, title = [None] * 5
			
		h, t = os.path.split(f)
		doc_id = Session.assign_uid()
		# FIELDS -- UID (I) | filename (CA) | path (CA) | mimetype (CA) | author (CA) | creator (CA) | producer (CA) | subject (CA) | title (CA)
		Session.main.ui.ls_documents.append([doc_id,    # UID (I)
                                             t,         # filename (CA)
                                             h,         # path (CA)
                                             mime,	    # mimetype (CA)
                                             author,    # author (CA)
                                             creator,   # creator (CA)
                                             producer,  # producer (CA)
                                             subject,   # subject (CA)
                                             title])    # title (CA)
		 
		Session.doc_crop[doc_id] = [[[0] * 4 for i in range(5)], 1, 2]
		Session.doc_temp[doc_id] = None
	
	def add_tag(self, tag):
		self.ui.txt_pattern.set_text(''.join([self.ui.txt_pattern.get_text(), tag]))
	
	def ask_proceed_with_action(self):
		if Session.has_changed and Session.filename:
			ans = message.yesnocancel(self, "There are unsaved changes. Would you like to save the current session before?")
			if ans == gtk.RESPONSE_YES:
				if self.on_mnu_file_save_activate(widget, data) != gtk.RESPONSE_OK: return False
			elif ans == gtk.RESPONSE_CANCEL:
				return False
		elif not Session.filename and Session.has_changed:
			ans = message.yesnocancel(self, "There are unsaved changes. Would you like to save the current session?")
			if ans == gtk.RESPONSE_YES:
				if self.on_mnu_file_save_activate(widget, data) != gtk.RESPONSE_OK: return
			elif ans == gtk.RESPONSE_CANCEL:
				return False
				
		return True	
	
	def concatenate(self):
		out_dlg = gtk.FileChooserDialog("Specify the file name of the output document", self,
										 buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
										            gtk.STOCK_OK, gtk.RESPONSE_OK),
										 action = gtk.FILE_CHOOSER_ACTION_SAVE)
		
		resp = out_dlg.run()
		if resp == gtk.RESPONSE_OK:
			dest = out_dlg.get_filename()
			out_dlg.destroy()
			with ProgressDialog() as pd:
				pd.thread = ConcatenateThread(pd, [self, dest])
				resp = pd.start()
				if resp != gtk.RESPONSE_OK:
					pd.stop()
				Session.status('File "%s" successfully created.' % dest)
				if preferences['chk_open_cat']: self.view(dest)
		else:
			out_dlg.destroy()
	
	def doc_info_apply(self):
		widgets = ['author', 'creator', 'producer', 'subject', 'title']
			
		new_data = [getattr(self.ui, 'txt_%s' % widget).get_text() for widget in widgets]
		prev = Session._dump_doc_list()
		model, it = self.ui.tv_documents.get_selection().get_selected()
		
		i = 4
		for field in new_data:
			model.set_value(it, i, field)
			i += 1	

	def draw_rectangle(self, widget, rect, color):
		cr = widget.window.cairo_create()
		s = self.preview.get_zoom()
		w, h = self.ui.da_preview.get_size_request()
		cr.set_line_width(.8)
		cr.set_source_rgb(*color)
		cr.rectangle(s * rect[RECT.LEFT],
		             s * rect[RECT.TOP],
		             (w - self.preview.pad - s * (rect[RECT.LEFT] + rect[RECT.RIGHT])),
		             (h - self.preview.pad - s * (rect[RECT.TOP] + rect[RECT.BOTTOM])))
		cr.stroke()

	def duplicate_doc_list(self):
		new_list = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING)
		
		it = self.ui.ls_documents.get_iter_first()
		while it:
			row = []
			for i in range(self.ui.ls_documents.get_n_columns()):
				row.append(self.ui.ls_documents.get_value(it, i))
			new_list.append(row)
			it = self.ui.ls_documents.iter_next(it)
			
		return new_list
	
	def get_from_selected_row(self, *args):
		model, it = self.ui.tv_documents.get_selection().get_selected()
		return model.get(it, *args) if it else None

	def get_selected_active_filename(self, create_temp = False):
		uid, path, name = self.get_from_selected_row(DOC_COLUMN.UID, DOC_COLUMN.PATH, DOC_COLUMN.FILENAME)
		temp_file = Session.doc_temp.get(uid)
		
		if temp_file: return temp_file
		
		if not temp_file and create_temp: # WARNING: Maybe it'd be one context for each file
			temp_context = Session.temp_new('temp_copy')
			temp_file = temp_context.new_file('.pdf')
			shutil.copyfile('/'.join((path, name)), temp_file)
			Session.doc_temp[uid] = temp_file
			filename = temp_file
		else:
			filename = '/'.join((path, name))
			
		return filename
	
	def get_selected_path(self):
		model, it = self.ui.tv_documents.get_selection().get_selected()
		return model.get_path(it) if it else None
	
	def process(self):
		with ProgressDialog() as pd:
			pd.thread = OverwriteThread(pd, self) if self.ui.rdo_overwrite.get_active() else SaveAsThread(pd, self)
			resp = pd.start()
			if resp != gtk.RESPONSE_OK:
				pd.stop()
			Session.status('Action successfully completed.')

	def remove(self):
		ts, paths = self.ui.tv_documents.get_selection().get_selected_rows()
		for path in paths: ts.remove(ts.get_iter(path))

	def set_at_selected_row(self, *args):
		model, it = self.ui.tv_documents.get_selection().get_selected()
		return model.set_value(it, *args) if it else None

	def set_preview_page(self, page):
		self.ui.txt_preview_page.set_text(str(page))
		self.update_bbox()

	def show_page_ranges(self, uid):	
		# PAGE RANGES
		ranges = Session.doc_ranges.get(uid)
		if ranges:
			is_include = ranges[0]
			ranges = ranges[1]
			what = 'custom'
			mod = False
		else:
			is_include = True
			ranges = ''
			what = 'all'
			mod = True
		self.ui.rdo_include.set_active(is_include)
		getattr(self.ui, 'rdo_%sclude' % ('in' if is_include else 'ex'), None).set_active(True)
		getattr(self.ui, 'rdo_%sclude_%s_pages' % ('in' if is_include else 'ex', what), None).set_active(mod or bool(ranges))
		getattr(self.ui, 'txt_%sclude_pages' % ('in' if is_include else 'ex'), None).set_text(ranges)

	def update_bbox(self):
		box = self.ui.cbo_box.get_active()
		new_values = Session.boxes[box][:]
		for widget, i in zip(BOX_WIDGETS, range(4)):
			getattr(self.ui, 'spb_%s' % widget, None).set_value(new_values[i])

	def view(self, filename):
		if preferences['rdo_viewer_default']:
			shell.gnome_open(filename)
		else:
			shell.execute('%s "%s"' % (preferences['txt_viewer_custom'], filename))

	def zoom_best_fit(self):
		factor = (self.ui.sw_preview.get_allocation().height - 10.) / self.preview.get_page_size()[1]
		if self.preview.get_zoom() != factor: self.preview.set_zoom(factor)
		
### END LOCAL HELPERS

### BEGIN SIGNAL HANDLERS

	def on_btn_add_author_clicked(self, widget, *args):
		self.add_tag('*author*')

	def on_btn_add_ext_clicked(self, widget, *args):
		self.add_tag('*ext*')

	def on_btn_add_filename_clicked(self, widget, *args):
		self.add_tag('*name*')
	
	def on_btn_add_numbering_clicked(self, widget, *args):
		n = self.ui.spb_digits.get_value()
		s = self.ui.spb_start.get_value()
		i = self.ui.spb_inc.get_value()
		self.add_tag('#{%d,%d,%d}' % (n, s, i))
	
	def on_btn_add_path_clicked(self, widget, *args):
		self.add_tag('*path*')

	def on_btn_add_title_clicked(self, widget, *args):
		self.add_tag('*title*')

	def on_btn_cat_browse_clicked(self, widget, data = None):
		out_dlg = gtk.FileChooserDialog("Specify the file name of the output document", self,
										 buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK,
													gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL),
										 action = gtk.FILE_CHOOSER_ACTION_SAVE)
		
		resp = out_dlg.run()
		if resp == gtk.RESPONSE_OK:
			self.ui.txt_cat_output.set_text(out_dlg.get_filename())
		out_dlg.destroy()

	def on_btn_dismiss_clicked(self, widget, *args):
		#self.on_btn_pattern_help_toggled(self.ui.btn_pattern_help, *args)
		self.ui.btn_pattern_help.set_active(False)

	def on_btn_doc_info_apply_clicked(self, widget, data = None):
		DocListCommand(self.doc_info_apply)

	def on_btn_pattern_help_toggled(self, widget, data = None):
		if widget.get_active():
			rect = txt = widget.get_allocation()
			#x, y = widget.translate_coordinates(widget.get_toplevel().get_toplevel(), 0, 0)
			x, y = widget.window.get_position()
			wx, wy = self.ui.win_pattern.get_position()
			w, h = self.ui.win_pattern.get_size()

			self.ui.win_pattern.move(wx - w/2, wy - h / 2 - rect.height)
			
			self.ui.win_pattern.show()
		else:
			self.ui.win_pattern.hide()

	def on_btn_perform_action_clicked(self, widget, data = None):
		# Determine the action
		if self.ui.rdo_cat.get_active():  # Concatenate
			self.concatenate()
		else:							  # Process separately
			self.process()

	def on_btn_preview_first_clicked(self, widget, *args):
		self.preview.show_page(0)
		self.set_preview_page(1)

	def on_btn_preview_last_clicked(self, widget, *args):
		page = self.preview.get_n_pages()
		self.preview.show_page(page - 1)
		self.set_preview_page(page)

	def on_btn_preview_next_clicked(self, widget, *args):
		page = self.preview.get_current_page()
		if page < self.preview.get_n_pages() - 1:
			self.preview.show_page(page + 1)
			self.set_preview_page(page + 2)

	def on_btn_preview_prev_clicked(self, widget, *args):
		page = self.preview.get_current_page()
		if page:
			self.preview.show_page(page - 1)
			self.set_preview_page(page)

	def on_btn_ranges_apply_clicked(self, widget, *args):
		#WARNING: The way this is implemented is particularly
		#         memory-consuming. This can surely be done way better
		#         but it'd probably require a bit more than just a
		#         moment's thought.
		uid, = self.get_from_selected_row(DOC_COLUMN.UID)
		
		is_include = self.ui.rdo_include.get_active()
		ranges = getattr(self.ui, 'txt_%sclude_pages' % ('in' if is_include else 'ex')).get_text()
		ApplyRanges(uid, is_include, ranges)
		
	def on_cbo_box_changed(self, widget, *args):
		for widget, value in zip(BOX_WIDGETS, Session.boxes[widget.get_active()]):
			getattr(self.ui, 'spb_%s' % widget).set_value(value)
		if getattr(self, 'preview', None): self.preview.show_page()

	def on_chk_crop_global_toggled(self, widget, *args):
		self.on_tv_documents_cursor_changed(self.ui.tv_documents, *args)
		if widget.get_active():
			self.ui.btn_crop_apply.set_sensitive(False)

	def on_chk_doc_info_global_toggled(self, widget, *args):
		self.on_tv_documents_cursor_changed(self.ui.tv_documents, *args)
		if widget.get_active():
			self.ui.btn_doc_info_apply.set_sensitive(False)

	def on_da_preview_expose_event(self, widget, event, *args):
		# Update the page preview (if the preview expander is expanded
		# and a document is actually selected)
		if getattr(self, 'preview', None):
			if self.ui.tb_zoom_best_fit.get_active(): self.zoom_best_fit()
			self.preview.expose()
					
		box = self.ui.cbo_box.get_active()
		color = gtk.gdk.Color(preferences["cb_box%d" % box])
		self.draw_rectangle(widget, Session.boxes[box], (color.red_float, color.green_float, color.blue_float))

	def on_gtumbler_window_delete_event(self, widget, *args):
		return not self.ask_proceed_with_action()

	def on_ls_documents_row_changed(self, widget, path, it, *args):
		self.on_tv_documents_cursor_changed(widget, *args)

	def on_ls_documents_row_deleted(self, widget, path, *args):
		if getattr(Session, 'reorder_fingerprint', None) != None:
			Session.reorder_fingerprint.append('DEL')

		self.ui.mnu_edit_remove_all.set_sensitive(bool(widget.get_iter_first()))
		self.on_tv_documents_cursor_changed(widget, *args)

	def on_ls_documents_row_inserted(self, widget, path, *args):
		if getattr(Session, 'reorder_fingerprint', None) != None:
			Session.reorder_fingerprint.append('INS')

		self.ui.mnu_edit_remove_all.set_sensitive(True)
		
	def on_mnu_file_new_activate(self, widget, data = None):	
		if self.ask_proceed_with_action(): Session.new()

	def on_mnu_file_open_activate(self, widget, data = None):
		if not self.ask_proceed_with_action(): return
		
		fc = gtk.FileChooserDialog("Select a gtumbler Session File (*.gts)", self, \
		                           gtk.FILE_CHOOSER_ACTION_OPEN, \
		                           ("Cancel",  gtk.RESPONSE_CANCEL, "OK", gtk.RESPONSE_OK))
		gts_filter = gtk.FileFilter()
		gts_filter.add_pattern('*.gts')
		gts_filter.set_name('gtumbler Session File (*.gts)')
		fc.add_filter(gts_filter)
		resp = fc.run()
		if resp == gtk.RESPONSE_OK:
			filename = fc.get_filename()
			if not Session.open(filename):
				message.error(self, "The specified file is not compatible with the current version of gtumbler.")
				return
		fc.destroy()

	def on_mnu_file_save_activate(self, widget, data = None):
		if Session.filename:
			Session.save()
		else:
			return self.on_mnu_file_save_as_activate(widget, data)

	def on_mnu_file_save_as_activate(self, widget, data = None):
		fc = gtk.FileChooserDialog("Save current session", self, \
		                           gtk.FILE_CHOOSER_ACTION_SAVE, \
		                           ("Cancel", gtk.RESPONSE_CANCEL,"OK", gtk.RESPONSE_OK))
		fc.set_do_overwrite_confirmation(True)
		gts_filter = gtk.FileFilter()
		gts_filter.add_pattern('*.gts')
		gts_filter.set_name('gtumbler Session File (*.gts)')
		fc.add_filter(gts_filter)
		resp = fc.run()
		if resp == gtk.RESPONSE_OK:
			Session.save(fc.get_filename())
		fc.destroy()
		return resp

	def on_mnu_edit_redo_activate(self, widget, data = None):
		self.on_tb_redo_clicked(widget, data)

	def on_mnu_edit_remove_activate(self, widget, data = None):
		self.on_tb_remove_clicked(widget, data)

	def on_mnu_edit_remove_all_activate(self, widget, data = None):
		DocListCommand(self.ui.ls_documents.clear)

	def on_mnu_edit_undo_activate(self, widget, data = None):
		self.on_tb_undo_clicked(widget, data)

	def on_rdo_cat_toggled(self, widget, data = None):
		self.ui.vb_separate.set_sensitive(not widget.get_active())

	def on_rdo_include_toggled(self, widget, data = None):
		self.ui.alg_include.set_sensitive(widget.get_active())
		self.ui.alg_exclude.set_sensitive(not widget.get_active())

	def on_rdo_overwrite_toggled(self, widget, data = None):
		self.ui.vb_saveas.set_sensitive(not widget.get_active())

	def on_spb_box_value_changed(self, widget, *args):
		widgets = zip(BOX_WIDGETS, range(4))
		
		box = self.ui.cbo_box.get_active()
		for widget, i in widgets:
			Session.boxes[box][i] = getattr(self.ui, 'spb_%s' % widget).get_value()
		
		self.preview.show_page()
		
	def on_btn_crop_apply_clicked(self, widget, *args):
		if ApplyCrop() == gtk.RESPONSE_OK:						
			Session.status('Action successfully completed.')	
			
			n = self.preview.get_current_page()
			filename = self.get_selected_active_filename()
			self.preview = DocumentPreview(filename, self.ui.da_preview)
			self.preview.show_page(n)
			self.update_bbox()		
		
	def on_tb_add_clicked(self, widget, data = None):
		##TODO: Figure out what the hell is wrong with this procedure!!!!
		open_dialog = gtk.FileChooserDialog("Select a document", self,
											buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK,
													   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
		open_dialog.set_select_multiple(True)
		doc_filter = gtk.FileFilter()
		for mime_type in mimes.CAT_MIMES: doc_filter.add_mime_type(mime_type)
		open_dialog.add_filter(doc_filter)
		
		resp = open_dialog.run()
		pd = None
		if resp == gtk.RESPONSE_OK:
			files = open_dialog.get_filenames()
			open_dialog.hide()
			prev = Session._dump_doc_list()
			with ProgressDialog() as pd:
				pd.thread = FileListThread(pd, [self, files])
				resp = pd.start()
				if resp != gtk.RESPONSE_OK:
					pd.stop()
				else:
					DocListCommand(previous=prev)
		else:
			open_dialog.hide()
	
	def on_tb_add_folder_activate(self, widget, data = None):
		doc_filter = gtk.FileFilter()
		for mime_type in mimes.CAT_MIMES: doc_filter.add_mime_type(mime_type)
		self.ui.fc_add_folder.add_filter(doc_filter)

		resp = self.ui.fc_add_folder.run()
		if resp == gtk.RESPONSE_OK:
			d = self.ui.fc_add_folder.get_filename()
			self.ui.fc_add_folder.hide()
			prev = Session._dump_doc_list()
			with ProgressDialog() as pd:
				pd.thread = DirectoryCrawlerThread(pd, [self, d, self.ui.chk_recursive.get_active()])
				resp = pd.start()
				if resp != gtk.RESPONSE_OK:
					pd.stop()
				else:
					DocListCommand(previous=prev)
		else:
			self.ui.fc_add_folder.hide()

	def on_tb_edit_clicked(self, widget, *args):
		filename = self.get_selected_active_filename(preferences['chk_create_copy'])
		editor = preferences['cbo_editor']
		args = preferences['txt_tool_args']
		if editor == 0:				# Xournal
			shell.xournal(args, filename)
		elif editor == 1:			# PDFedit
			shell.pdfedit(args, filename)
		elif editor == -1:			# Custom editor
			app = preferences['txt_other_tool']
			shell.execute('%s %s "%s"' % (app, args, filename))
		else: # TODO: No editor selected. Bring Preferences Dialog up
			pass

	def on_tb_redo_clicked(self, widget, data = None):
		Session.redo()
	
	def on_tb_remove_clicked(self, widget, data = None):
		DocListCommand(self.remove)

	def on_tb_undo_clicked(self, widget, data = None):
		Session.undo()
	
	def on_tb_view_clicked(self, widget, *args):
		self.view(self.get_selected_active_filename())
	
	def on_tb_zoom_best_fit_toggled(self, widget, *args):
		if widget.get_active(): self.zoom_best_fit()
		
	def on_tb_zoom_in_clicked(self, widget, *args):
		self.ui.tb_zoom_best_fit.set_active(False)
		self.preview.zoom_in(0.10)
		
	def on_tb_zoom_out_clicked(self, widget, *args):
		self.ui.tb_zoom_best_fit.set_active(False)
		self.preview.zoom_out(0.10)

	def on_tv_documents_cursor_changed(self, widget, data = None):
		#TODO: The structure of this procedure can be made more sensible
		#      if each step is defined in a single procedure on its own.
		#      Then the global structure would be:
		#      * retrieve current uid
		#      * if uid is valid then update interface
		#      * else hide and stuff
		
		# The following widgets will be sensitive only when a document is
		# selected
		widgets = [('tb_remove', None),
		           ('mnu_edit_remove', None),
		           ('btn_doc_info_apply', None),
		           ('btn_crop_apply', None),
		           ('vb_doc_info', self.ui.chk_doc_info_global.get_active()),
		           ('vb_crop', self.ui.chk_crop_global.get_active()),
		           ('tb_edit', None),
		           ('tb_view', None),
		           ('exp_preview', None),
		           ('exp_ranges', None)]
		
		# INFO
		info = self.get_from_selected_row(4, 5, 6, 7, 8)
		for widget, cond in widgets: getattr(self.ui, widget).set_sensitive(bool(info) or bool(cond))
		if not info: info = [''] * 5
		
		widgets = [self.ui.txt_author, self.ui.txt_creator, self.ui.txt_producer, self.ui.txt_subject, self.ui.txt_title]
		for i in range(len(info)): widgets[i].set_text(info[i] if info[i] else '')
		
		# PREVIEW
		data = self.get_from_selected_row(DOC_COLUMN.UID)
		uid = data[0] if data else None
		if data:
			filename = self.get_selected_active_filename()
			self.preview = DocumentPreview(filename, self.ui.da_preview)
			self.preview.show_page(0)
			self.ui.lbl_n_pages.set_text(' of %d ' % self.preview.get_n_pages())
			self.set_preview_page(1)
		else:
			self.ui.exp_preview.set_expanded(False)
			
		# PAGE RANGES
		if uid != None: self.show_page_ranges(uid)
		
	def on_tv_documents_drag_begin(self, drag_context, *args):
		Session.drag_dump = Session._dump_doc_list()
		Session.reorder_fingerprint = []

	def on_tv_documents_drag_end(self, drag_context, *args):
		## WARNING: This might be a bit slow
		if Session.reorder_fingerprint == ['INS', 'DEL']:
			current = Session._dump_doc_list()
			if current != Session.drag_dump:
				DocListCommand(previous = Session.drag_dump)
		Session.reorder_fingerprint = None
		
	def on_txt_preview_page_activate(self, widget, *args):
		try:
			new_page = int(widget.get_text())
			
			n_pages = self.preview.get_n_pages()
			
			if new_page < 0:
				new_page += self.preview.get_n_pages() + 1
			#if not new_page or new_page > self.preview.get_n_pages():
			
			# This is more human-friendly but probably slower
			if new_page not in range(1, n_pages + 1):
				raise ValueError
			
		except ValueError:
			widget.set_text(str(self.preview.get_current_page() + 1))
			return
			
		widget.set_text(str(new_page))
		self.preview.show_page(new_page - 1)
			
		
### END SIGNAL HANDLERS
