# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2012 Phoenix87 <phoenix1987@gmail.com>
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

# This is your preferences dialog.
#
# Define your preferences dictionary in the __init__.main() function.
# The widget names in the PreferencesTestProjectDialog.ui
# file need to correspond to the keys in the preferences dictionary.
#
# Each preference also need to be defined in the 'widget_methods' map below
# to show up in the dialog itself.  Provide three bits of information:
#  1) The first entry is the method on the widget that grabs a value from the
#     widget.
#  2) The second entry is the method on the widget that sets the widgets value
#      from a stored preference.
#  3) The third entry is a signal the widget will send when the contents have
#     been changed by the user. The preferences dictionary is always up to
# date and will signal the rest of the application about these changes.
# The values will be saved to desktopcouch when the application closes.
#
# TODO: replace widget_methods with your own values

widget_class = {
	'ColorButton'	: ['get_color_spec', 'set_color_spec', 'color-set'],
	'ComboBox'			: ['get_active', 'set_active', 'changed'],
	'Entry'				: ['get_text', 'set_text', 'key-release-event'],
	'ToggleButton'		: ['get_active', 'set_active', 'toggled'],
}

widget_methods = {
	'cb_box0'				: widget_class['ColorButton'],
	'cb_box1'				: widget_class['ColorButton'],
	'cb_box2'				: widget_class['ColorButton'],
	'cb_box3'				: widget_class['ColorButton'],
	'cb_box4'				: widget_class['ColorButton'],

    'cbo_editor'			: widget_class['ComboBox'],
    'cbo_compatibility'     : widget_class['ComboBox'],

    'chk_convert'			: widget_class['ToggleButton'],
    'chk_create_copy'		: widget_class['ToggleButton'],
    'chk_crop_create_copy'	: widget_class['ToggleButton'],
    'chk_display_path'		: widget_class['ToggleButton'],
    'chk_open_cat' 			: widget_class['ToggleButton'],
    
    'rdo_viewer_default'	: widget_class['ToggleButton'],
    'rdo_viewer_custom'		: widget_class['ToggleButton'],
    
    'txt_other_tool'		: widget_class['Entry'],
    'txt_tool_args'			: widget_class['Entry'],
    'txt_viewer_custom'		: widget_class['Entry'],
}

import gettext
from gettext import gettext as _
gettext.textdomain('gtumbler')

import logging
logger = logging.getLogger('gtumbler')

from gtumbler_lib.PreferencesDialog import PreferencesDialog

import session
Session = session.Session()

class PreferencesGtumblerDialog(PreferencesDialog):
    __gtype_name__ = "PreferencesGtumblerDialog"

    def finish_initializing(self, builder): # pylint: disable=E1002
        """Set up the preferences dialog"""
        super(PreferencesGtumblerDialog, self).finish_initializing(builder)

        # populate the dialog from the preferences dictionary
        # using the methods from widget_methods
        self.widget_methods = widget_methods
        
        self.ui.cbo_editor.set_property('active', 2)
        self.set_widgets_from_preferences() # pylint: disable=E1101

        # Code for other initialization actions should be added here.

### BEGIN SIGNAL HANDLERS

	def on_cbo_editor_changed(self, widget, *args):
		self.ui.txt_other_tool.set_visible(widget.get_active_text() == "Other...")
	
	def on_preferences_gtumbler_dialog_response(self, widget, *args):
		Session._update_window_title()
	
	def on_rdo_viewer_default_toggled(self, widget, *args):
		self.ui.txt_viewer_custom.set_sensitive(not widget.get_active())

### END SIGNAL HANDLERS
